from secrets import token_hex

from django.conf import settings

from monerorpc.authproxy import AuthServiceProxy


def verify_signature(address: str, challenge: str, signature: str) -> bool:
    """Makes a request to wallet RPC to verify address and signature."""
    protocol = settings.MONERO_WALLET_RPC_PROTOCOL
    host = settings.MONERO_WALLET_RPC_HOST
    user = settings.MONERO_WALLET_RPC_USER
    pwd = settings.MONERO_WALLET_RPC_PASS
    wallet_rpc = AuthServiceProxy(f"{protocol}://{user}:{pwd}@{host}/json_rpc")

    result = wallet_rpc.verify(
        {"data": challenge, "address": address, "signature": signature}
    )

    return result.get("good", False)


def generate_challenge():
    """Generates a new random challenge for the authentication."""
    return token_hex(8)



# -*- coding: utf-8 -*-
from django.utils.translation import gettext_lazy as _
from django.contrib.auth.views import LoginView
from django.contrib.auth import get_user_model
from django.views.generic import FormView
from django.forms.utils import ErrorList
from django.conf import settings
from django.urls import reverse

from monerorpc.authproxy import JSONRPCException

from .forms import SimpleSignUpForm, SimpleLoginForm
from .utils import verify_signature


class MoneroLoginView(LoginView):
    form_class = SimpleLoginForm


class MoneroSignUpView(FormView):
    template_name = "registration/signup.html"
    form_class = SimpleSignUpForm

    def get_form(self, form_class=None):
        return self.form_class(request=self.request, **self.get_form_kwargs())

    def form_valid(self, form):
        username = form.cleaned_data["username"]
        address = form.cleaned_data["address"]
        challenge = form.cleaned_data["challenge"]
        signature = form.cleaned_data["signature"]
        try:
            valid_sig = verify_signature(address, challenge, signature)
        except JSONRPCException:
            form._errors["__all__"] = ErrorList([_("Error connecting to daemon")])
            return self.form_invalid(form)

        if valid_sig:
            user = get_user_model().objects.create(username=username)
            user.address_set.create(address=address)
            return super().form_valid(form)
        else:
            form._errors["signature"] = ErrorList([_("Invalid signature")])
            return self.form_invalid(form)

    def get_success_url(self):
        return settings.LOGIN_REDIRECT_URL



# Generated by Django 2.2.5 on 2019-10-15 17:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_cryptolock.validators
import model_utils.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="Address",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="modified",
                    ),
                ),
                (
                    "address",
                    models.CharField(
                        max_length=106,
                        unique=True,
                        validators=[
                            django_cryptolock.validators.validate_monero_address
                        ],
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"verbose_name": "Address", "verbose_name_plural": "Addresses"},
        )
    ]



from django.contrib.auth.backends import ModelBackend
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext_lazy as _

from monerorpc.authproxy import AuthServiceProxy, JSONRPCException

from .models import Address
from .utils import verify_signature

User = get_user_model()


class MoneroAddressBackend(ModelBackend):
    """Custom Monero-Cryptolock authentication backend."""

    def authenticate(self, request, address=None, challenge=None, signature=None):
        """Validates the provided signature for the given address and challenge.

        This method currently relies on Wallet RPC access to verify the signature,
        in the future it should be done locally to be more reliable and more
        performant.
        """
        if not all([address, challenge, signature]):
            return None

        stored_address = (
            Address.objects.select_related("user").filter(address=address).first()
        )
        if not stored_address:
            return None

        try:
            is_valid = verify_signature(address, challenge, signature)
        except JSONRPCException:
            raise PermissionDenied(_("Error while validating signature"))

        if is_valid:
            return stored_address.user

        return None



# -*- coding: utf-8 -*-

from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from model_utils.models import TimeStampedModel

from .validators import validate_monero_address


class Address(TimeStampedModel):
    """Addresses that belong to a given user account."""

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    address = models.CharField(
        max_length=106, validators=[validate_monero_address], unique=True
    )

    class Meta:
        """Meta definition for Address."""

        verbose_name = _("Address")
        verbose_name_plural = _("Addresses")

    def __str__(self):
        """Unicode representation of Address."""
        return self.address
